/*
 **********************************************************************************
 *  Copyright (C) 2010-2011  Massachusetts Institute of Technology
 *  Copyright (C) 2010-2011  Yuan Tang <yuantang@csail.mit.edu>
 * 		                     Charles E. Leiserson <cel@mit.edu>
 * 	 
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *   Suggestsions:                  yuantang@csail.mit.edu
 *   Bugs:                          yuantang@csail.mit.edu
 *
 *********************************************************************************
 */

/* Test bench - 2D heat equation, Periodic version */
#include <cstdio>
#include <cstddef>
#include <iostream>
#include <cstdlib>
#include <sys/time.h>
#include <cmath>

#include <pochoir.hpp>

using namespace std;
#define TIMES 1
#define N_RANK 3
#define TOLERANCE (1e-6)

void check_result(int t, int j, int i, int k, double a, double b)
{
	if (abs(a - b) < TOLERANCE) {
//		printf("a(%d, %d, %d) == b(%d, %d, %d) == %f : passed!\n", t, j, i, t, j, i, a);
	} else {
		printf("a(%d, %d, %d) = %f, b(%d, %d, %d) = %f : FAILED!\n", t, j, i, a, t, j, i, b);
	}

}

Pochoir_Boundary_3D(aperiodic_3D, arr, t, i, j, k)
    return 0;
Pochoir_Boundary_End

Pochoir_Boundary_3D(periodic_3D, arr, t, i, j, k)
    const int arr_size_1 = arr.size(1);
    const int arr_size_0 = arr.size(0);
    const int arr_size_2 = arr.size(2);

    int new_i = (i >= arr_size_1) ? (i - arr_size_1) : (i < 0 ? i + arr_size_1 : i);
    int new_j = (j >= arr_size_0) ? (j - arr_size_0) : (j < 0 ? j + arr_size_0 : j);
    int new_k = (k >= arr_size_2) ? (k - arr_size_2) : (k < 0 ? k + arr_size_2 : k);

    /* we use arr.get(...) instead of arr(...) to implement different boundary
     * checking strategy: In arr(...), if off-boundary access occurs, we call
     * boundary function to supply a value; in arr.get(...), if off-boundary 
     * access occurs, we will print the off-boundary access and quit!
     */
return arr.get(t, new_i, new_j, new_k);
    // return arr.get(t, -1, -1);
Pochoir_Boundary_End

int main(int argc, char * argv[])
{
	const int BASE = 1024;
	int t;
	struct timeval start, end;
    int N_SIZE = 0, T_SIZE = 0;

    if (argc < 3) {
        printf("argc < 3, quit! \n");
        exit(1);
    }
    N_SIZE = StrToInt(argv[1]);
    T_SIZE = StrToInt(argv[2]);
    printf("N_SIZE = %d, T_SIZE = %d\n", N_SIZE, T_SIZE);
    Pochoir_Shape_3D heat_shape_3D[] = {{0, 0, 0, 0}, {-1, 1, 0, 0}, {-1, -1, 0, 0}, {-1, 0, 0, 0}, {-1, 0, 0, -1}, {-1, 0, 0, 1}, {-1, 0, 1, 0}, {-1, 0, -1, 0}};
    Pochoir<N_RANK> heat_3D(heat_shape_3D);
    Pochoir_Array<double, N_RANK> a(N_SIZE, N_SIZE, N_SIZE), b(N_SIZE, N_SIZE, N_SIZE);
    a.Register_Boundary(periodic_3D);
    heat_3D.Register_Array(a);

    b.Register_Shape(heat_shape_3D);
    b.Register_Boundary(periodic_3D);

    /* Now we can only access the Pochoir_Array after Register_Array,
     * or Register_Shape with the array, because we rely on the shape
     * to provide the depth of toggle array!!! 
     */
	for (int i = 0; i < N_SIZE; ++i) {
	for (int j = 0; j < N_SIZE; ++j) {
	for(int k = 0; k < N_SIZE; ++k) {
	    a(0, i, j, k) = 1.0 * (rand() % BASE); 
	    a(1, i, j, k) = 0; 
	    b(0, i, j, k) = a(0, i, j, k);
	    b(1, i, j, k) = 0;
	} } }

	cout << "a(T+1, J, I) = 0.125 * (a(T, J+1, I) - 2.0 * a(T, J, I) + a(T, J-1, I)) + 0.125 * (a(T, J, I+1) - 2.0 * a(T, J, I) + a(T, J, I-1)) + a(T, J, I)" << endl;
	Pochoir_Kernel_3D(heat_3D_fn, t, i, j, k)
	   a(t, i, j, k) = 
           0.125 * (a(t-1, i+1, j, k) - 2.0 * a(t-1, i, j, k) + a(t-1, i-1, j, k)) 
         + 0.125 * (a(t-1, i, j+1, k) - 2.0 * a(t-1, i, j, k) + a(t-1, i, j-1, k)) 
         + 0.125 * (a(t-1, i, j, k+1) - 2.0 * a(t-1, i, j, k) + a(t-1, i, j, k-1))
         + a(t-1, i, j, k);
    Pochoir_Kernel_End

	gettimeofday(&start, 0);
    for (int times = 0; times < TIMES; ++times) {
        heat_3D.Run(T_SIZE, heat_3D_fn);
    }
	gettimeofday(&end, 0);
	std::cout << "Pochoir ET: consumed time :" << 1.0e3 * tdiff(&end, &start)/TIMES << "ms" << std::endl;

#if 0
	gettimeofday(&start, 0);
    for (int times = 0; times < TIMES; ++times) {
	for (int t = 0; t < T_SIZE; ++t) {
    cilk_for (int i = 0; i < N_SIZE; ++i) {
	for (int j = 0; j < N_SIZE; ++j) {
	    for(int k = 0; k < N_SIZE; ++k) {
		b(t+1, i, j, k) = 
		    0.125 * (b(t, i+1, j, k) - 2.0 * b(t, i, j, k) + b(t, i-1, j, k)) 
		    + 0.125 * (b(t, i, j+1, k) - 2.0 * b(t, i, j, k) + b(t, i, j-1, k)) 
		    + 0.125 * (b(t, i, j, k+1) - 2.0 * b(t, i, j, k) + b(t, i, j, k-1))
		    + b(t, i, j, k); }}}}
	    }
	gettimeofday(&end, 0);
	std::cout << "Naive Loop: consumed time :" << 1.0e3 * tdiff(&end, &start)/TIMES << "ms" << std::endl;

	t = T_SIZE;
	for (int i = 0; i < N_SIZE; ++i) {
	for (int j = 0; j < N_SIZE; ++j) {
	    for(int k = 0; k < N_SIZE; ++k) {
	    check_result(t, i, j, k, a.interior(t, i, j), b.interior(t, i, j));
	    } } }
#endif

	return 0;
}
